r"""
# Example usage
```
required_packages = [
    ("numpy", "1.20.3"),  # Specify version
    ("pandas", None),     # Latest version
]

dynamic_package_import(required_packages)
```
"""

import importlib.metadata
import subprocess
import sys
from typing import List, Tuple

def install_or_upgrade_package(package: str, version: str = None) -> None:
    """
    Installs or upgrades a Python package using pip via Popen.
    
    :param package: The name of the package to install or upgrade.
    :param version: Optional. The version of the package to install.
    """
    # Check if the package is already installed
    try:
        dist = importlib.metadata.distribution(package)
        current_version = dist.version
        print(f"{package} is already installed at version {current_version}.")
        if version and current_version != version:
            print(f"Upgrading {package} to version {version}.")
            _execute_pip_command(f"install {package}=={version}")
        else:
            print("No upgrade needed.")
    except importlib.metadata.PackageNotFoundError:
        print(f"{package} is not installed. Installing...")
        _execute_pip_command(f"install {package}{'==' + version if version else ''}")


def _execute_pip_command(command: str) -> None:
    """
    Executes a pip command using subprocess.Popen.
    
    :param command: The pip command to execute (e.g., 'install numpy').
    """
    pip_cmd = [sys.executable, "-m", "pip", *command.split()]
    with subprocess.Popen(pip_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True) as proc:
        stdout, stderr = proc.communicate()
        if proc.returncode != 0:
            print(f"Error executing pip command: {command}")
            print(f"Error details: {stderr}")
        else:
            print(f"Pip command '{command}' executed successfully.")


def dynamic_package_import(required_packages: List[Tuple[str, str]]) -> None:
    """
    Checks for the presence of required packages and installs/upgrades them if necessary.
    
    :param required_packages: A list of tuples, where each tuple contains a package name and optionally a version.
    """
    for package, version in required_packages:
        install_or_upgrade_package(package, version)